//
// Scilab ( http://www.scilab.org/ ) - This file is part of Scilab
// Copyright (C) 2021 - Stephane MOTTELET
//
// This file is hereby licensed under the terms of the GNU GPL v3.0,
// For more information, see the COPYING file which you should have received
//
//--------------------------------------------------------------------------

#include "OdeManager.hxx"
#include "SUNDIALSBridge.hxx"

static std::map<std::wstring, int> intCbState{{L"init",-1}, {L"step",0}, {L"done",1}};

int SUNDIALSEvent(realtype t, N_Vector N_VectorY, realtype *pdblOut, void *pManager)
{
    OdeManager *manager = static_cast<OdeManager *>(pManager);
    int iNbRealEq = manager->getNRealEq();

    dynlibFunPtr pFunc = manager->getEntryPointFunction(OdeManager::EVENTS);
    dynlibFunPtr pFuncWithPar = manager->getEntryPointFunctionWithPar(OdeManager::EVENTS);
    if (pFunc != NULL)
    {
        // Computation is made by user C/C++/Fortran dynamically lindked entrypoint
        int iNbEvents = manager->getNbEvents();
        ((SUNDIALSDynEvent)pFunc)(&iNbRealEq, &t, NV_DATA_S(N_VectorY), &iNbEvents, pdblOut, NULL, NULL);
    }
    else if (pFuncWithPar != NULL)
    {
        int iNbEvents = manager->getNbEvents();
        types::Double *pDblPar = manager->getParameters(OdeManager::EVENTS)[0]->getAs<types::Double>();
        ((SUNDIALSDynEventWithPar)pFuncWithPar)(&iNbRealEq, &t, NV_DATA_S(N_VectorY), &iNbEvents, pdblOut, NULL, NULL, pDblPar->get());        
    }
    else
    {
        manager->computeFunction(t, NV_DATA_S(N_VectorY), NULL, OdeManager::EVENTS, pdblOut);        
    }
    return 0;
}

int SUNDIALSImplEvent(realtype t, N_Vector N_VectorY, N_Vector N_VectorYp, realtype *pdblOut, void *pManager)
{
    OdeManager *manager = static_cast<OdeManager *>(pManager);
    int iNbRealEq = manager->getNRealEq();

    dynlibFunPtr pFunc = manager->getEntryPointFunction(OdeManager::EVENTS);
    dynlibFunPtr pFuncWithPar = manager->getEntryPointFunctionWithPar(OdeManager::EVENTS);
    if (pFunc != NULL)
    {
        // Computation is made by user C/C++/Fortran dynamically lindked entrypoint
        int iNbEvents = manager->getNbEvents();
        ((SUNDIALSDynImplEvent)pFunc)(&iNbRealEq, &t, NV_DATA_S(N_VectorY), NV_DATA_S(N_VectorYp), &iNbEvents, pdblOut, NULL, NULL);
    }
    else if (pFuncWithPar != NULL)
    {
        int iNbEvents = manager->getNbEvents();
        types::Double *pDblPar = manager->getParameters(OdeManager::EVENTS)[0]->getAs<types::Double>();
        ((SUNDIALSDynImplEventWithPar)pFuncWithPar)(&iNbRealEq, &t, NV_DATA_S(N_VectorY), NV_DATA_S(N_VectorYp), &iNbEvents, pdblOut, NULL, NULL, pDblPar->get());        
    }
    else
    {
        manager->computeFunction(t, NV_DATA_S(N_VectorY), NV_DATA_S(N_VectorYp), OdeManager::EVENTS, pdblOut);        
    }
    return 0;
}

int SUNDIALSIntCb(realtype t, N_Vector N_VectorY, N_Vector N_VectorYp, wchar_t *pwstFlag, void *pManager)
{
    int iRet = 0;
    OdeManager *manager = static_cast<OdeManager *>(pManager);
    int iNbRealEq = manager->getNRealEq();

    dynlibFunPtr pFunc = manager->getEntryPointFunction(OdeManager::INTCB);
    dynlibFunPtr pFuncWithPar = manager->getEntryPointFunctionWithPar(OdeManager::INTCB);
    if (pFunc != NULL)
    {
        // Computation is made by DLL entrypoint
        int iState = intCbState[pwstFlag];
        if (manager->isDAE())
        {
            iRet = ((SUNDIALSDynImplIntCb)pFunc)(&iNbRealEq, &t, NV_DATA_S(N_VectorY),  NV_DATA_S(N_VectorYp), &iState);                        
        }
        else
        {
            iRet = ((SUNDIALSDynIntCb)pFunc)(&iNbRealEq, &t, NV_DATA_S(N_VectorY), &iState);            
        }
    }
    else if (pFuncWithPar != NULL)
    {
        // Computation is made by DLL entrypoint with double array extra parameter
        types::Double *pDblPar = manager->getParameters(OdeManager::INTCB)[0]->getAs<types::Double>();
        int iState = intCbState[pwstFlag];
        if (manager->isDAE())
        {
            iRet = ((SUNDIALSDynImplIntCbWithPar)pFuncWithPar)(&iNbRealEq, &t, NV_DATA_S(N_VectorY), NV_DATA_S(N_VectorYp), &iState, pDblPar->get());            
        }
        else
        {
            iRet = ((SUNDIALSDynIntCbWithPar)pFuncWithPar)(&iNbRealEq, &t, NV_DATA_S(N_VectorY), &iState, pDblPar->get());            
        }
    }
    else
    {
        iRet = (int) manager->intermediateCallback(t, NV_DATA_S(N_VectorY), manager->isDAE() ? NV_DATA_S(N_VectorYp) : NULL, pwstFlag);
    }
    
    return iRet;
}

void SUNDIALSErrHandler(int error_code, const char *module, const char *function, char *msg, void *pManager)
{
    OdeManager *manager = static_cast<OdeManager *>(pManager);
    manager->errHandler(error_code, module, function, msg);
}
