//
// Scilab ( http://www.scilab.org/ ) - This file is part of Scilab
// Copyright (C) 2021 - Stephane MOTTELET
//
// This file is hereby licensed under the terms of the GNU GPL v3.0,
// For more information, see the COPYING file which you should have received
//
//--------------------------------------------------------------------------

#include "OdeManager.hxx"
#include "SUNDIALSBridge.hxx"

void OdeManager::solve()
{
    solverReturnCode iFlag;
    int iLastOrder;
    int iNbErrors = 0; 
    int iNbSteps = 0;
    int iFirstStep = 0;
    double dblTime = 0;
    double dblCurrTime;
    double dblStep;
    double dblFinalTime;
    bool bTerminalEvent = false;
    char errorMsg[256];
    OdeManager *prevManager = getPreviousManager();
    
    SUNDIALSIntCb(m_dblT0, m_N_VectorY, m_N_VectorYp, L"init", (void *)this);
    if ((m_pDblTSpan->getSize() == 2 && std::isnan(m_dblOptT0)) || m_iRetCount <= 1)        
    {
        std::vector<double> newY;
        std::vector<double> newYp;
        
        newY.reserve(m_iNbRealEq);
        if (m_isDAE)
        {
            newYp.reserve(m_iNbRealEq);            
        }
        /*
            Case where we compute solution only at method steps guided only by precision (RTOL, ATOL) requirements.
            This occurs when tspan is of the kind [t0 t1] or in the case of sol = method() syntax, i.e. when we yield
            a solution structure with a pointer to the OdeManager instance for further custom evaluation by interpolation 
            or time extension.
        */
        if (setStopTime())
        {
            sprintf(errorMsg,"setStopTime error\n");             
            throw ast::InternalError(errorMsg);
        }
        dblFinalTime = m_pDblTSpan->get(m_pDblTSpan->getSize()-1);
        
        // Copy (t0,y0) only if not extending a previous solution
        if (m_odeIsExtension == false)
        {
            // TODO: reserve space for vectors
            m_indexNS.push_back(0);
            m_vecYOut.push_back(std::vector<double>(NV_DATA_S(m_N_VectorY),NV_DATA_S(m_N_VectorY) + m_iNbRealEq));
            m_dblVecTOut.push_back(m_dblT0);
            if (m_isDAE)
            {
                m_vecYpOut.push_back(std::vector<double>(NV_DATA_S(m_N_VectorYp),NV_DATA_S(m_N_VectorYp) + m_iNbRealEq));                
            }
        }
        else
        {
            // new values will be appended to previous vectors
            // TODO: reserve space for vectors
            m_indexNS = prevManager->m_indexNS;
            m_vecYOut = prevManager->m_vecYOut;
            m_dblVecTOut = prevManager->m_dblVecTOut; // Beware, m_dblVecTOut has T0 hence its size is m_iVecOrder.size()+1
            m_dblVecCurrTime = prevManager->m_dblVecCurrTime;
            m_iVecOrder = prevManager->m_iVecOrder;
            // events
            m_dblVecYEvent = prevManager->m_dblVecYEvent;
            m_dblVecIndexEvent = prevManager->m_dblVecIndexEvent;
            m_dblVecEventTime = prevManager->m_dblVecEventTime;
            // index of further steps takes into account previous values
            iFirstStep = prevManager->m_iVecOrder.size();
            if (m_isDAE)
            {
                m_vecYpOut = prevManager->m_vecYpOut;
                m_dblVecYpEvent = prevManager->m_dblVecYpEvent;                
            }
        }    

        iNbSteps = iFirstStep;
        

        for (; m_dblT0 != dblFinalTime;)
        {
            // Solver internal step
            iFlag = doStep(dblFinalTime, m_N_VectorY, m_N_VectorYp, &dblTime, ODE_ONE_STEP);
            if (iFlag == ODE_SUCCESS || iFlag == ODE_TSTOP_RETURN || iFlag == ODE_ROOT_RETURN)
            {
                // Should not be necessary
                if (dblTime == m_dblVecTOut.back())
                {
                    errHandler(ODE_WARNING, NULL, NULL, NULL);
                    break;
                }

                newY.assign(NV_DATA_S(m_N_VectorY), NV_DATA_S(m_N_VectorY) + m_iNbRealEq);
                if (m_isDAE)
                {
                    newYp.assign(NV_DATA_S(m_N_VectorYp), NV_DATA_S(m_N_VectorYp) + m_iNbRealEq);                    
                }
                if (getLastOrder(&iLastOrder))
                {
                    sprintf(errorMsg,"getLastOrder error\n");             
                    throw ast::InternalError(errorMsg);   
                }
                if (getCurrentTime(&dblCurrTime))
                {
                    sprintf(errorMsg,"getCurrentTime error\n");             
                    throw ast::InternalError(errorMsg);                       
                }
                // Refine by interpolation if requested and [t,y,...] output 
                if (m_iNRefine > 1 && m_iRetCount > 1)
                {
                    N_Vector N_VectorYRefine = N_VNew_Serial(m_iNbRealEq);
                    double dblStepRef = (dblTime-m_dblVecTOut.back())/(double)m_iNRefine;
                    double dblRefTime = m_dblVecTOut.back()+dblStepRef;
                    for (int i=1; i < m_iNRefine; i++)
                    {
                        if (getDky(dblRefTime, 0, N_VectorYRefine))
                        {
                            sprintf(errorMsg,"getDky error\n");             
                            throw ast::InternalError(errorMsg);                                                   
                        }
                        std::vector<double> newY (NV_DATA_S(N_VectorYRefine), NV_DATA_S(N_VectorYRefine) + m_iNbRealEq);
                        m_vecYOut.push_back(newY);
                        if (m_isDAE)
                        {
                            if (getDky(dblRefTime, 1, N_VectorYRefine))
                            {
                                sprintf(errorMsg,"getDky error\n");             
                                throw ast::InternalError(errorMsg);                                                   
                            }
                            std::vector<double> newYp (NV_DATA_S(N_VectorYRefine), NV_DATA_S(N_VectorYRefine) + m_iNbRealEq);
                            m_vecYpOut.push_back(newYp);                            
                        }
                        m_dblVecTOut.push_back(dblRefTime);
                        dblRefTime +=dblStepRef;
                    }
                    N_VDestroy_Serial(N_VectorYRefine);
                }

                if (iFlag == ODE_ROOT_RETURN)
                {                    
                    int *piRootsFound = new int(m_iNbEvents);
                    
                    if (getRootInfo(piRootsFound))
                    {
                        delete piRootsFound;
                        sprintf(errorMsg,"getRootInfo erro\n");             
                        throw ast::InternalError(errorMsg);                                                                           
                    }
                    
                    std::vector<double> newIndex (piRootsFound, piRootsFound + m_iNbEvents);
                    delete piRootsFound;
                    
                    // Check if one event is a terminal event
                    if (m_iVecEventIsTerminal.size() > 0)
                    {
                        for (int i=0; i < m_iNbEvents; i++)
                        {
                            bTerminalEvent |= (newIndex[i] !=0 && m_iVecEventIsTerminal[i] == 1);
                        }
                        // bTerminalEvent == true will stop integration 
                    }
                    m_dblVecEventTime.push_back(dblTime);
                    m_dblVecYEvent.push_back(newY);
                    if (m_isDAE)
                    {
                        m_dblVecYpEvent.push_back(newYp);                        
                    }
                    m_dblVecIndexEvent.push_back(newIndex);
                }
                m_vecYOut.push_back(newY);
                if (m_isDAE)
                {
                    m_vecYpOut.push_back(newYp);                    
                }
                m_dblVecCurrTime.push_back(dblCurrTime);
                m_dblVecTOut.push_back(dblTime);
                m_iVecOrder.push_back(iLastOrder);
                iNbSteps++;
                
                if (m_iRetCount <= 1)
                { 
                    saveContinuousExtensionVectors(); 
                }
                
                if (SUNDIALSIntCb(dblTime, m_N_VectorY, m_N_VectorYp, L"step", (void *)this))                    
                {
                    break;
                }
                if (bTerminalEvent == true || iFlag == ODE_TSTOP_RETURN)
                {
                    break;
                }
            }
            else
            {
                // Errors/Warnings not trapped by SUNDIALS ErrorFunc
                errHandler(iFlag, NULL, NULL, NULL);
            }
        }

        if (m_iRetCount == 1) // Solution structure output
        {
            double *pdblNS;
            // Size may vary among solvers: m_iNbRealEq for CVODE and m_iNbRealEq+1 for IDA 
            int iSize = pDblDerOutList.front().size();

            m_pDblNSMatrices = new types::Double(iSize, m_indexNS.back());
            // Check if we are extending a previous solution
            if (m_odeIsExtension)
            {
                types::Double *pDblPrevNSMatrices = prevManager->m_pDblNSMatrices;

                // copy previous NS matrices
                std::copy(pDblPrevNSMatrices->get(), pDblPrevNSMatrices->get()+pDblPrevNSMatrices->getSize(), m_pDblNSMatrices->get());
                pdblNS = m_pDblNSMatrices->get() + pDblPrevNSMatrices->getSize();
            }
            else
            {
                pdblNS = m_pDblNSMatrices->get();
            }

            for (int i = iFirstStep; i < iNbSteps; i++)
            {
                for (int j = 0; j < m_iVecOrder[i]+1; j++)
                {
                    double *pdbl = pDblDerOutList.front().data();            
                    std::copy(pdbl, pdbl+iSize, pdblNS);
                    
                    pDblDerOutList.pop_front();
                    pdblNS += iSize;
                }
            }
        }
    }
    else // Case where we compute solution at prescribed user timesteps, with method stepping and dense output interpolation
    {
        double *pdblYout = NULL;
        double *pdblYoutImg = NULL;
        double *pdblYpout = NULL;
        double *pdblYpoutImg = NULL;
        m_pDblTOut = m_pDblTSpan->clone();
        m_pDblYOut = createYOut(m_iNbEq, m_pDblTSpan->getSize());
        pdblYout = m_pDblYOut->get();
        pdblYoutImg = m_pDblYOut->getImg();
        
        if (m_isDAE)
        {
            m_pDblYpOut = createYOut(m_iNbEq, m_pDblTSpan->getSize());            
            pdblYpout = m_pDblYpOut->get();
            pdblYpoutImg = m_pDblYpOut->getImg();
        }

        if (m_dblT0 == m_pDblTSpan->get(0))
        {
            std::copy(NV_DATA_S(m_N_VectorY),NV_DATA_S(m_N_VectorY)+m_iNbEq, pdblYout);
            pdblYout += m_iNbEq;
            if (m_odeIsComplex)
            {
                std::copy(NV_DATA_S(m_N_VectorY)+m_iNbEq,NV_DATA_S(m_N_VectorY)+2*m_iNbEq, pdblYoutImg);                
                pdblYoutImg += m_iNbEq;
            }
            if (m_isDAE)
            {
                std::copy(NV_DATA_S(m_N_VectorYp),NV_DATA_S(m_N_VectorYp)+m_iNbEq, pdblYpout);
                pdblYpout += m_iNbEq;
                if (m_odeIsComplex)
                {
                    std::copy(NV_DATA_S(m_N_VectorYp)+m_iNbEq, NV_DATA_S(m_N_VectorYp)+2*m_iNbEq, pdblYpoutImg);                
                    pdblYpoutImg += m_iNbEq;
                }                
            }
        }

        for (int i = (int) m_dblT0 == m_pDblTSpan->get(0); i<m_pDblTSpan->getSize();)
        {
            iFlag = doStep(m_pDblTSpan->get(i), m_N_VectorY, m_N_VectorYp, &dblTime, ODE_NORMAL);

            if (iFlag == ODE_ROOT_RETURN)
            {
                int *piRootsFound = new int(m_iNbEvents);
                if (getRootInfo(piRootsFound))
                {
                    delete piRootsFound;
                    sprintf(errorMsg,"getRootInfo erro\n");             
                    throw ast::InternalError(errorMsg);                                                                           
                }
                std::vector<double> newIndex (piRootsFound, piRootsFound + m_iNbEvents);
                delete piRootsFound;

                m_dblVecEventTime.push_back(dblTime);
                std::vector<double> newY (NV_DATA_S(m_N_VectorY), NV_DATA_S(m_N_VectorY) + m_iNbRealEq);
                m_dblVecYEvent.push_back(newY);
                if (m_isDAE)
                {
                    std::vector<double> newYp (NV_DATA_S(m_N_VectorYp), NV_DATA_S(m_N_VectorYp) + m_iNbRealEq);
                    m_dblVecYEvent.push_back(newYp);                    
                }
                m_dblVecIndexEvent.push_back(newIndex);

                // Check if one event is a terminal event
                if (m_iVecEventIsTerminal.size() > 0)
                {
                    for (int j=0; j < m_iNbEvents; j++)
                    {
                        bTerminalEvent |= (newIndex[j] !=0 && m_iVecEventIsTerminal[j] == 1);
                    }
                    if (bTerminalEvent == true)
                    {
                        // Stop integration at time of event, precisely (don't finish at next user step)
                        m_pDblTOut->set(i,dblTime);
                        
                        std::copy(NV_DATA_S(m_N_VectorY),NV_DATA_S(m_N_VectorY)+m_iNbEq,pdblYout);
                        pdblYout += m_iNbEq;
                        if (m_odeIsComplex)
                        {
                            std::copy(NV_DATA_S(m_N_VectorY)+m_iNbEq,NV_DATA_S(m_N_VectorY)+2*m_iNbEq,pdblYout);
                            pdblYoutImg += m_iNbEq;
                        }
                        if (m_isDAE)
                        {
                            std::copy(NV_DATA_S(m_N_VectorYp),NV_DATA_S(m_N_VectorYp)+m_iNbEq,pdblYpout);
                            pdblYpout += m_iNbEq;
                            if (m_odeIsComplex)
                            {
                                std::copy(NV_DATA_S(m_N_VectorYp)+m_iNbEq,NV_DATA_S(m_N_VectorYp)+2*m_iNbEq,pdblYpout);
                                pdblYpoutImg += m_iNbEq;
                            }                            
                        }
                        
                        i++;
                        // do a resize of pdblYout and m_pDblTOut
                        m_pDblTOut->resize(1,i);
                        int iDims =  m_pDblYOut->getDims();
                        int *piDimsArray = m_pDblYOut->getDimsArray();
                        int *piNewDimsArray = new int(iDims);
                        std::copy(piDimsArray, piDimsArray+iDims, piNewDimsArray);
                        piNewDimsArray[iDims-1] = i;
                        m_pDblYOut->resize(piNewDimsArray, iDims);
                        if (m_isDAE)
                        {
                            m_pDblYpOut->resize(piNewDimsArray, iDims);                            
                        }
                        delete piNewDimsArray;
                        break;
                    }
                }
            }
            else if (iFlag == ODE_SUCCESS)
            {
                std::copy(NV_DATA_S(m_N_VectorY),NV_DATA_S(m_N_VectorY)+m_iNbEq,pdblYout);
                pdblYout += m_iNbEq;
                if (m_odeIsComplex)
                {
                    std::copy(NV_DATA_S(m_N_VectorY)+m_iNbEq,NV_DATA_S(m_N_VectorY)+2*m_iNbEq,pdblYoutImg);
                    pdblYoutImg += m_iNbEq;
                }
                if (m_isDAE)
                {
                    std::copy(NV_DATA_S(m_N_VectorYp),NV_DATA_S(m_N_VectorYp)+m_iNbEq,pdblYpout);
                    pdblYpout += m_iNbEq;
                    if (m_odeIsComplex)
                    {
                        std::copy(NV_DATA_S(m_N_VectorYp)+m_iNbEq,NV_DATA_S(m_N_VectorYp)+2*m_iNbEq,pdblYpoutImg);
                        pdblYpoutImg += m_iNbEq;
                    }                    
                }
                i++;
            }
            else
            {
                // Errors/Warnings not trapped by SUNDIALS ErrorFunc
                errHandler(iFlag, NULL, NULL, NULL);
            }
            if (SUNDIALSIntCb(dblTime, m_N_VectorY, m_N_VectorYp, L"step", (void *)this))                    
            {
                break;
            }
        }
    }
    
    SUNDIALSIntCb(dblTime, m_N_VectorY, m_N_VectorYp, L"done", (void *)this);
    // save last current step for future extension
    getCurrentStep(&m_dblCurrentStep);
}