// =============================================================================
// Scilab ( http://www.scilab.org/ ) - This file is part of Scilab
//
// Copyright (C) 2021 - Stéphane Mottelet
//
// This file must be used under the terms of GPL License
// This source file is licensed as described in the file COPYING, which
// you should have received as part of this distribution.  The terms
// are also available atsol
// https://www.gnu.org/licenses/gpl-3.0.txt

//
function dydt = vdp(t,y,mu)
    dydt = [y(2,:)
            mu*(1-y(1,:).*y(1,:)).*y(2,:)-y(1,:)];
end

function out = crhs(t,y)
    out = 10*exp(2*%i*%pi*t)*y;
endfunction

function out=vdpJac(t,y,mu)
    out = [0 1
          -2*mu*y(1)*y(2)-1 mu*(1-y(1)*y(1))];
endfunction

function dySdt = vdpsens(t,yS,mu)
    y=yS(:,1);
    S=yS(:,2);
    dySdt = [vdp(t,y,mu) vdpJac(t,y,mu)*S+[0;(1-y(1)^2)*y(2)]];
end

function [out,term,dir] = myevent(t,y,term,dir)
    out  = y(1)-1.7;
endfunction


function xdot = matmul(t,x,A)
    xdot=A*x;
endfunction

function Xdot=ric(t,X,A,B,C)
    Xdot=A'*X+X*A-X'*B*X+C
endfunction

function acc = bounce(t,y,g)
    acc = [y(2); -g]
endfunction

function [out,term,dir] = bounce_ev(t,y)
    out = y(1);
    term = 1;
    dir = -1;
endfunction

function dydt=sir(t,y,bet,gam,N)
    dydt=[-bet/N*y(1)*y(2)
           bet/N*y(1)*y(2)-gam*y(2)  
           gam*y(2)];
endfunction

function out = sir_ev(t,y,bet,gam,N)
    out = bet/N*y(1)*y(2)-gam*y(2)
endfunction

function out = scicallback(t,y,flag,stats)
    out = %f
    if flag == "step"
        mprintf("%s : hlast=%g\n",flag,stats.hlast)
    end
endfunction

mu = 1;
y0 = [2;1];

//BASIC
[t,y] = cvode(list(vdp,mu), 0:0.1:10, y0);
assert_checkalmostequal(y(:,$),[-1.6992622;-1.6126613]);
//VERY STIFF
[t,y] = cvode(list(vdp,1000), [0 3000], [2;1], method="bdf");
assert_checkalmostequal(y(:,$),[-1.50512748918; 0.00118938688]);

[t,y] = cvode(list(vdp,mu), 0:0.1:10, y0, method="bdf");
assert_checkalmostequal(y(:,$),[-1.7010025;-1.6116379]);

[t,y] = cvode(list(vdp,mu), 0:0.1:10, y0, method="bdf",rtol=1e-10);
[ts,ys] = cvode(list(vdp,mu), [0 10], y0, method="bdf",rtol=1e-10);
assert_checkalmostequal(y(:,$),ys(:,$),1e-7);

// DYNAMIC LINK RHS
rhscode=['void dynrhs(int *n,double *t,double *y,double *ydot)'
       '{'
       'ydot[0] = y[1];'
       'ydot[1] = (1-y[0]*y[0])*y[1]-y[0];'
       '}']
rhscodewithpar=['void dynrhspar(int *n,double *t,double *y,double *ydot,double *mu)'
       '{'
       'ydot[0] = y[1];'
       'ydot[1] = mu[0]*(1-y[0]*y[0])*y[1]-y[0];'
       '}']
jaccode=['void dynjac(int *n,double *t,double *y,double *ydot, double *J)'
       '{'
       'J[0] = 0; J[1] = -2*y[0]*y[1]-1;'
       'J[2] = 1.0; J[3] = 1-y[0]*y[0];'
       '}'];
jaccodewithpar=['void dynjacpar(int *n,double *t,double *y,double *ydot, double *J, double *mu)'
       '{'
       'J[0]=0; J[1]=-2.0*mu[0]*y[0]*y[1]-1.0;'
       'J[2]=1.0; J[3]=mu[0]*(1.0-y[0]*y[0]);'
       '}'];
cd TMPDIR
mputl(rhscode,TMPDIR+'/dynrhs.c') //create the C file
ilib_for_link('dynrhs','dynrhs.c',[],'c',[]);//compile
exec("TMPDIR/loader.sce",-1)
[tg,yg]=cvode('dynrhs',[0 10], y0);
[t,y]=cvode(list(vdp,1),[0 10],y0);
assert_checkequal(y,yg)

// DYNAMIC LINK RHS WITH PARAMETER
mputl(rhscodewithpar,TMPDIR+'/dynrhspar.c') //create the C file
ilib_for_link('dynrhspar','dynrhspar.c',[],'c',[]);//compile
exec("TMPDIR/loader.sce",-1)
[tg,yg]=cvode(list('dynrhspar',1),[0 10], y0);
assert_checkequal(y,yg)

// DYNAMIC LINK RHS WITH PARAMETER AND JACOBIAN
mputl(jaccode,TMPDIR+'/dynjac.c') //create the C file
ilib_for_link('dynjac','dynjac.c',[],'c',[]);//compile
exec("TMPDIR/loader.sce",-1)
[t,y]=cvode(list(vdp,1),[0 10],y0,  method="bdf");
[tg,yg]=cvode('dynrhs',[0 10], y0, jacobian="dynjac", method="bdf");
assert_checkalmostequal(y,yg,1e-3)
[tg1,yg1]=cvode('dynrhs',[0 10],y0, jacobian=list(vdpJac,1), method="bdf");
assert_checkequal(yg,yg1)

// EVENTS
[t,y,te,ye,ie] = cvode(list(vdp,mu), [0,10], y0, method="bdf",jacobian=list(vdpJac,mu),events=list(myevent,0,0));
assert_checkalmostequal(ye(1,:),[1.7,1.7 1.7])
assert_checkalmostequal(te,[ 1.21746186   6.668978609   7.860696911])
//
[t,y,te,ye,ie] = cvode(list(vdp,mu), [0,10], y0, method="bdf",jacobian=list(vdpJac,mu),events=list(myevent,0,-1));
assert_checkalmostequal(ye(1,:),[1.7, 1.7])
assert_checkalmostequal(te,[ 1.21746186    7.860696911])
//
[t,y,te,ye,ie] = cvode(list(vdp,mu), [0,10], y0, method="bdf",jacobian=list(vdpJac,mu),events=list(myevent,0,-1));
assert_checkalmostequal(ye(1,:),[1.7, 1.7])
assert_checkalmostequal(te,[ 1.21746186    7.860696911])
//
[t,y,te,ye,ie] = cvode(list(vdp,mu), [0,10], y0, method="bdf",jacobian=list(vdpJac,mu),events=list(myevent,1,1));
assert_checkequal(te, t($))
assert_checkalmostequal(ye(1,:),[1.7])
assert_checkalmostequal(te,[6.668978609])
//
[t,y,te,ye,ie] = cvode(list(vdp,mu), [0,10], y0, method="bdf",jacobian=list(vdpJac,mu),events=list(myevent,0,0));
sol = cvode(list(vdp,mu), [0,10], y0, method="bdf", jacobian=list(vdpJac,mu), events=list(myevent,0,0));
assert_checkequal(t,sol.t)
assert_checkequal(y,sol.y)
assert_checkequal(te,sol.te)
assert_checkequal(ye,sol.ye)
assert_checkequal(ie,sol.ie)
assert_checkalmostequal(sol(sol.t),(sol.y))

// EXTEND SOLUTION
sol = cvode(list(vdp,mu), [0 10], y0, method="bdf");
solext = cvode(sol, 20)
assert_checkequal(size(sol.t),[1,159])
assert_checkequal(size(solext.t),[1,304])
assert_checkequal(sol(sol.t),solext(sol.t))

//  EXTEND SOLUTION WITH EVENT
sol2 = cvode(list(vdp,mu), [0,10], y0, method="bdf",events=list(myevent,1,0));
sol2ext = cvode(sol2, 20)
assert_checkequal(sol2(sol2.t),sol2ext(sol2.t))

// MATRIX ODE
[t,y] = cvode(list(matmul,[1 1;0 2]), 1, eye(2,2), t0=0);
sol3 = cvode(list(matmul,[1 1;0 2]), 1, eye(2,2), t0=0);
assert_checkequal(size(sol3.y),[2,2,19])
[t,E] = cvode(list(matmul,[1 1;0 2]), 1, eye(2,2), t0=0);
assert_checkalmostequal(E,[ 2.718281767 4.670543306; 0,7.388825073]);

// MATRIX ODE
A=[1,1;0,2]; B=[1,0;0,1]; C=[1,0;0,1];
[t,X]=cvode(list(ric,[1,1;0,2],[1,0;0,1],[1,0;0,1]), %pi, eye(A),t0=0)
assert_checkalmostequal(X,[  2.272713109   0.615508769; 0.615508769   4.41906588 ]);

// COMPLEX ODE
solc = cvode(crhs,[0 5],1);
assert_checkalmostequal(solc.y(:,$),complex( 1.00009192413402314692,-0.000078229167544612501))
solcext1=cvode(solc,6)
solcext2=cvode(solc,6,y0=1)
assert_checkalmostequal(solcext2(5+10*%eps),1,0,100*%eps)
solcext3=cvode(solc,6,y0=%i)
assert_checkalmostequal(solcext3(5+10*%eps),%i,0,100*%eps)

// SENSITIVITY WITH COMPLEX STEP
h = 1e-200;
mu=1
[tcs,ycs] = cvode(list(vdp,complex(mu,h)), [0,10], y0, rtol=1e-14);
scs = imag(ycs)/h;
[tcs,sens] = cvode(list(vdpsens,mu), tcs, [y0 [0;0]], rtol=1e-14);
sens = squeeze(sens(:,2,:))
assert_checktrue(max(abs(sens-scs)) < 3e-4)

// EXTEND SOLUTION + EVENTS, 
sol = cvode(list(bounce,9.81), [0 10], [1;0], events=bounce_ev);
for i=1:100
    yini = sol.y(:,$).*[1;-0.9];
    sol = cvode(sol, 10, y0=yini);
    if sol.te($)-sol.te($-1) < 1e-3
        break
    end
end
assert_checkalmostequal(sol.te($),8.55768);
t = linspace(0, sol.t($), 10000);
//clf
//plot(sol.te, sol.ye(1,:), 'or', t, sol(t,1))

// REFINE+EVENTS
//clf
//drawlater
yini = [1;0];
te = 0;
while %t
    [t,y,tev,ye] = cvode(list(bounce,9.81), [te($) 100], yini, events=bounce_ev, refine=16);
    if tev==[]
        break;
    end
//    plot(t,y(1,:),te,ye(1),'o')
    yini = y(:,$).*[1;-0.8];
    te = [te;tev]
end
assert_checkalmostequal(te($), 4.0588273);
//drawnow

// SIR MODEL
N=60e6;
gam=1/40;
bet=0.2;
y0=[N-1;1;0];
[t,y,te,ye]=cvode(list(sir,bet,gam,N),[0 400],y0,events=sir_ev);
assert_checkalmostequal(te, 114.561646320925021314);
assert_checkalmostequal(ye,[7499999.999998786486685
   36903765.29698479920626
   15596234.70301640406251
]);
//clf
//plot(t,y,te,ye(2),'or');
//legend "Susceptible" "Infected" "Recovered"

//HEAT EQUATION
function [dv] = f_chaleur(t,v,h,lambda,c,rho)
	f=x>1/4 & x<1/3;
  dv = (f-lambda/h^2*( -[0;v(1:$-1)] + 2*v -[v(2:$);0] ))/c/rho;
endfunction
L=1; N=1000;
lambda = 1;
c=200;
rho=7893;
d=0.02;
tf=300;
section=%pi*d^2/4;
rhoLin=rho*section;
dx = L/N; x = linspace(dx,L-dx,N-1)';
v0=zeros(N-1,1);
t=[0 logspace(0,3,19)]// vecteur des temps
tic;
[t,v] = cvode(list(f_chaleur,dx,lambda,c,rhoLin),t,v0,method="bdf",rtol=1e-5,atol=1e-7);
t1=toc();
tic;
v2=ode("stiff",v0,0,t,1e-5,1e-7,list(f_chaleur,dx,lambda,c,rhoLin));
t2 = toc();
assert_checktrue(max(abs(v-v2)) < 1e-6);
assert_checktrue(t1/t2 < 0.5);

//POSITIVENESS
function out=fp(t,y)
    out = -y;
endfunction

[t,y] = cvode(fp,[0 70],1,atol=1e-12)
assert_checktrue(or(y < 0))
[t,y] = cvode(fp,[0 70],1,atol=1e-12,positive=1)
assert_checktrue(y > 0)

//SCILAB ERRORS
function fe1(t,y)
    out = -y;
endfunction
msg = msprintf(_("%s: Wrong number of output argument(s): %d expected."),"fe1",1);
assert_checkerror("cvode(fe1,[0 70],1)",msg)

function out = fe2(t,y)
    undefined_symbol
endfunction
msg = msprintf(_("Undefined variable: %s\n"),"undefined_symbol");
assert_checkerror("cvode(fe2,[0 70],1)",msg)

//function out=fe3(t,y)
//    out=%inf
//endfunction
//assert_checkerror("cvode(fe3,[0 70],1)","cvode: At t = 0 and h = 0, the corrector convergence test failed repeatedly or with |h| = hmin.")

// SINGULAR SOLUTION AT T=1
function out = fe4(t,y)
    out = y^2;
endfunction
[t,y] = cvode(fe4,[0 2],1);
assert_checkalmostequal(t($),0.999630242)

// CALLBACK
cbcode=['int dyncb(int *n,double *t,double *y, int *flag)'
       '{'
       'sciprint(""flag=%d\n"",*flag);'
       'return 0;'
       '}'];
cd TMPDIR
mputl(cbcode,TMPDIR+'/dyncb.c') //create the C file
ilib_for_link('dyncb','dyncb.c',[],'c',[]);//compile
exec("TMPDIR/loader.sce",-1)
[t,y] = cvode(fe4,[0 0.5],1,intcb="dyncb");
[t,y] = cvode(fe4,[0 1],1,intcb=scicallback);

// DYNAMIC EVENTS
eventcode=['void dynevent(int *n,double *t,double *y, int *ng, double *g, int *term, int *dir)'
       '{ '
       'if (*ng == 0) {'
       '    *ng = 1; '
       '} else if (g == 0) {'
       '    term[0] = 1; dir[0] = 1;'
       '} else {'
       '    g[0] = y[0]-1.7;'
       '}}'];
cd TMPDIR
mputl(eventcode,TMPDIR+'/dynevent.c') //create the C file
ilib_for_link('dynevent','dynevent.c',[],'c',[]);//compile
exec("TMPDIR/loader.sce",-1)

[t,y,te,ye] = cvode(list(vdp,1), [0 10], [2;1], method="bdf",events="dynevent");
assert_checkalmostequal(y(1,$),1.7)

// DYNAMIC EVENTS WITH PARAMETER
eventcodewithpar=['void dyneventpar(int *n,double *t,double *y, int *ng, double *g, int *term, int *dir, double *par)'
       '{ '
       'if (*ng == 0) {'
       '    *ng = 1; '
       '} else if (g == 0) {'
       '    term[0] = 1; dir[0] = 1;'
       '} else {'
       '    g[0] = y[0]-par[0];'
       '}}'];
cd TMPDIR
mputl(eventcodewithpar,TMPDIR+'/dyneventpar.c') //create the C file
ilib_for_link('dyneventpar','dyneventpar.c',[],'c',[]);//compile
exec("TMPDIR/loader.sce",-1)

[t,y,te,ye] = cvode(list(vdp,1), [0 10], [2;1], method="bdf",events=list("dyneventpar",1));
assert_checkalmostequal(ye(1),1)

[t,y,te,ye] = cvode(list(vdp,1), [0 10], [2;1], method="bdf",events=list("dyneventpar",1.234));
assert_checkalmostequal(ye(1),1.234)
